/**
 * @file
 * Provides Splidebox, splide within lightbox.
 */

/* global define, module */
(function (root, factory) {

  'use strict';

  var ns = 'Splidebox';
  var db = root.dBlazy;

  // Inspired by https://github.com/addyosmani/memoize.js/blob/master/memoize.js
  if (typeof define === 'function' && define.amd) {
    // AMD. Register as an anonymous module.
    define([ns, db, Drupal, root], factory);
  }
  else if (typeof exports === 'object') {
    // Node. Does not work with strict CommonJS, but only CommonJS-like
    // environments that support module.exports, like Node.
    module.exports = factory(ns, db, Drupal, root);
  }
  else {
    // Browser globals (root is window).
    root[ns] = factory(ns, db, Drupal, root);
  }

}((this || module || {}), function (ns, $, Drupal, _win) {

  'use strict';

  var _doc = document;
  var _base = 'splide';
  var _sBase = '.' + _base;
  var _name = _base + 'box';
  var _nick = 'sbox';
  var _isNick = 'is-' + _nick;
  var _isMounted = _isNick + '-mounted';
  var _isMounting = _isNick + '-mounting';
  var _isPlaying = _isNick + '-playing';
  var _isActive = _isNick + '-active';
  var _isUmounting = _isNick + '-umounting';
  var _isFullscreened = _isNick + '-fullscreen';
  var _isPaginated = _isNick + '-paginated';
  var _isMoved = _isNick + '-moved';
  var _isZoomable = 'is-zoomable';
  var _isCentered = 'is-centered';
  var _ariaHidden = 'aria-hidden';
  var _isInitial = false;
  var _classSplideList = _base + '__list';
  var _dragObj = _name + '__item';
  var _data = 'data-';
  var _dataName = _data + _name;
  var _media = 'media';
  var _dataMedia = _data + _nick + '-' + _media;
  var _dataProvider = _data + _nick + '-provider';
  var _img = 'img';
  var _src = 'src';
  var _srcset = 'srcset';
  var _sizes = 'sizes';
  var _html = 'html';
  var _template = 'template';
  var _playable = 'playable';
  var _audio = 'audio';
  var _iframe = 'iframe';
  var _video = 'video';
  var _isPlayable = _isNick + '-' + _playable;
  var _isVisible = _isNick + '-visible';
  var _isCaptioned = 'is-captioned'; // @todo rename it: _isNick + '-captioned';
  var _isAjaxed = _isNick + '-ajaxed';
  var _addClass = 'addClass';
  var _removeClass = 'removeClass';
  var _trigger;
  var fnSplide = false;
  var fnFullscreen = false;
  var fnMedia = false;
  var fnZoom = false;
  var _sanitizer = $.sanitizer;
  var _overlay = 'overlay';
  var _click = 'click';
  var _width = 'width';
  var _height = 'height';
  var _transitionend = 'transitionend.' + _nick;
  var _suffixes = [
    _overlay,
    'content',
    'caption',
    'counter',
    'closer',
    'fullscreen'
  ];
  var _defaults = {
    captionPos: _overlay,
    layout: 'bottom',
    trigger: '[' + _dataName + '-trigger]',
    skin: '',
    fsIconOn: '',
    template: '',
    dataSplide: {},
    init: false
  };

  var _itemDefaults = {
    root: null,
    width: 0,
    height: 0,
    i: 0,
    html: '',
    alt: '',
    caption: '',
    msrc: '',
    src: '',
    srcset: '',
    type: '',
    rect: {},
    ajax: ''
  };

  var fn = Splidebox.prototype;
  fn.constructor = Splidebox;

  /**
   * Defines Splidebox constructor.
   *
   * @param {HTMLElement} el
   *   The Splidebox HTML element.
   * @param {object} opts
   *   The Splidebox options.
   *
   * @return {Splidebox}
   *   The Splidebox instance.
   *
   * @namespace
   */
  function Splidebox(el, opts) {
    var me = $.extend({}, fn, this);

    me.name = ns;
    me.options = $.extend({}, _defaults, opts);
    me.gallery = el;

    // DOM ready fix.
    setTimeout(function () {
      init(me);
    });

    return me;
  }

  // Public methods.
  fn.toData = function (el) {
    var me = this;
    var opts = me.options;
    var data = $.parse($.attr(el, 'data-b-media data-media'));
    var div = $.find(el, '.' + _media);
    var href = $.attr(el, 'href');
    var url = $.attr(el, _data + 'box-url', href, true);
    var img = $.find(el, _img);
    var check = el.nextElementSibling;
    var caption = $.hasClass(check, 'visually-hidden') ? check : null;
    var rect = $.rect(el);
    var ajax = $.attr(el, _data + 'b-ajax');
    var html = '';
    var type = data.boxType || '';
    var display = type;
    var alt = $.attr(img, 'alt');

    if ([_iframe, _audio, _video].includes(type)) {
      display = _playable;
    }
    if (ajax) {
      display = _html;
    }

    var item = {
      element: el,
      width: parseInt(data.width, 0),
      height: parseInt(data.height, 0),
      i: me.getIndex(el),
      alt: _sanitizer.sanitize(alt),
      caption: $.isElm(caption) ? caption.innerHTML : '',
      msrc: $.attr(div, 'data-b-thumb data-thumb'),
      provider: data.provider,
      type: type,
      rect: rect,
      isPicture: false,
      ajax: ajax,
      ajaxOnly: (ajax && opts.ajaxOnly) || false,
      display: display
    };

    if (_html in data) {
      html = data.html;

      if (data.encoded) {
        html = atob(html);
      }

      // Only supports non-picture, unfortunately.
      var isResimage = $.contains(html, _srcset);
      var isPicture = $.contains(html, '<picture');
      item.isPicture = isPicture;

      if (isResimage && !isPicture) {
        item.srcset = extractAttr(html, _srcset);
        item.src = extractAttr(html, _src);
        item.sizes = extractAttr(html, _sizes);
      }
      else {
        if ([_audio, _video].includes(type)) {
          item.src = url;
        }
        // Local video and picture.
        item.html = me.htmlWrapper({
          html: html,
          data: item
        });
      }
    }
    else if (type === _iframe) {
      item.html = Drupal.theme('blazyMedia', {
        el: el
      });
    }
    else {
      item.src = url;
    }

    return $.extend({}, _itemDefaults, item);
  };

  fn.addSlide = function (data) {
    var me = this;
    var el = me.list;
    var slide = me.toSlide(data);

    if ($.isElm(el)) {
      append(el, slide);

      data.slide = slide;
      $.trigger(me.slider, _name + '.added', {
        data: data
      });
    }
  };

  fn.toSlide = function (data) {
    var me = this;
    var content;

    if (data.html) {
      content = data.html;
    }
    else {
      if (data.src) {
        var img = preload(data);
        content = img.outerHTML;
      }
    }

    var html = me.slideWrapper({
      html: content,
      data: data
    });

    var el = createElement(_template, html);
    var picture = $.find(el, 'picture');

    if ($.isElm(picture) && !data.ajax) {
      var dim = $.windowSize();
      var pic = $.find(picture, 'img');
      $.addClass(picture, _isCentered);
      $.addClass(pic, _name + '__img ' + _isZoomable);
      $.attr(pic, 'loading', 'lazy');
      $.attr(pic, _width, (dim.width - 32));
      $.attr(pic, _height, (dim.height - 22));
    }
    return el;
  };

  /**
   * Gets the current clicked item index.
   *
   * @param {HTMLElement} link
   *   The link HTML element.
   *
   * @return {Int}
   *   The current clicked link index.
   */
  fn.getIndex = function (link) {
    var me = this;
    var i = 0;
    if (me.items.length) {
      $.each(me.items, function (data, idx) {
        if (data.element && data.element === link) {
          i = idx;
          return i;
        }
      });
    }

    return i;
  };

  fn.open = function (e) {
    var me = this;
    e.preventDefault();
    e.stopPropagation();

    _isInitial = true;
    resetInitials(me);

    // With a mix of (non-)lightboxed contents: image, video, Facebook,
    // Instagram, etc., some may not always be lightboxed, so filter em out.
    var target = e.target;
    var parent = $.closest(target, 'a');
    var link = $.attr(parent, 'href') ? parent : $.closest(target, _trigger);
    var triggers = [];
    var pos;

    if ($.isStr(_trigger)) {
      triggers = $.findAll(me.gallery, _trigger);
    }

    me.count = triggers.length;
    me.triggers = triggers;

    if (me.count) {
      $.each(triggers, function (el, i) {
        var clone = $.closest(el, _sBase + '__slide--clone');
        if (!$.isElm(clone)) {
          var data = me.toData(el);
          data.i = i;
          me.items.push(data);
        }
      });
    }

    pos = me.getIndex(link);
    launch.call(me, pos);
  };

  fn.close = function () {
    var me = this;

    if ($.hasClass(me.root, _isFullscreened)) {
      if (fnFullscreen) {
        me.fullscreen.click();
      }

      return false;
    }

    me.destroy();
  };

  fn.destroy = function () {
    var me = this;
    var root = me.root;

    $.addClass(root, _isUmounting);

    var box = me.toBox();

    if (!box || !box.isValid || box.isPlayable || box.isAjax) {
      detach(me);
      return;
    }

    var img = box.img.el;
    var tn = box.tn;
    var z = tn.z;

    me.zoom('off');

    img.style.transform =
      'translate3d(' +
      tn.x +
      'px, ' +
      tn.y +
      'px, 0) scale3d(' +
      z +
      ', ' +
      z +
      ', 1)';

    var reset = function (e) {
      img.style.transform = 'translate3d(0,0,0) scale3d(1,1,1)';
      detach(me);
      $.off(img, _transitionend, reset);
    };

    $.on(img, _transitionend, reset);

    $.off(_win, 'blazy.mediaPlaying', onPlaying.bind(me));
    $.off(_win, 'blazy.mediaStopped', onStopped.bind(me));

    // Failsafe with potential screwed-up transition.
    setTimeout(function () {
      detach(me);
    }, 600);
  };

  fn.toBox = function () {
    var me = this;
    var data = me.currData || me.items[me.index];

    if ($.isUnd(data)) {
      return {};
    }

    var link = data.element;
    var slide = $.isEmpty(me.curr) ? data.slide : me.curr.slide;
    var img = $.find(slide, _img);
    var timg = $.find(link, _img);
    var ih = 0;
    var iw = 0;
    var il = 0;
    var it = 0;
    var th = 0;
    var tw = 0;
    var tx = 0;
    var ty = 0;
    var tz = 0;
    var tl = 0;
    var tt = 0;
    var ws = $.windowSize();
    var ww = ws.width;
    var wh = ws.height;
    var iRect = {};
    var tRect = {};
    var isAjax = data.ajax !== '';
    var isIframe = data.type === _iframe;
    var isVideo = data.type === _video;

    if (!$.isElm(img)) {
      var video = $.find(slide, _video);
      var poster = $.attr(video, 'poster');

      if (poster) {
        img = video;
      }
    }

    if ($.isElm(img)) {
      iRect = $.rect(img);
      ih = $.outerHeight(img);
      iw = $.outerWidth(img);
      il = iRect.left;
      it = iRect.top;
    }

    if ($.isElm(timg)) {
      var ox = 1;
      var oy = 1;
      tRect = $.rect(timg);
      th = $.outerHeight(timg);
      tw = $.outerWidth(timg);
      tl = tRect.left;
      tt = tRect.top;

      var aox = Math.abs(ox);
      var aoy = Math.abs(oy);
      var tminX = (ox > 0 ? 1 - aox : aox) * ww + (ox * ww) / 2;
      var tminY = (oy > 0 ? 1 - aoy : aoy) * wh + (oy * wh) / 2;

      tx = tl + tw / 2 - tminX;
      ty = tt + th / 2 - tminY;
      tz = tw / iw;
    }

    var box = {
      hasImg: $.isElm(img),
      hasTn: $.isElm(timg),
      isAjax: isAjax,
      isValid: $.isElm(img) && $.isElm(timg),
      isIframe: isIframe,
      isVideo: isVideo,
      isPlayable: isIframe || isVideo,
      ws: ws,
      img: {
        el: img,
        height: ih,
        width: iw,
        left: il,
        top: it,
        rect: iRect
      },
      tn: {
        el: timg,
        src: data.msrc || (timg && timg.src),
        x: tx,
        y: ty,
        z: tz,
        height: th,
        width: tw,
        left: tl,
        top: tt,
        rect: tRect
      }
    };
    me.currBox = box;
    return box;
  };

  fn.zoom = function (e) {
    if (fnZoom) {
      fnZoom[e]();
    }
  };

  // Private Methods
  function buildRoot() {
    var me = this;
    var root = me.root = createElement(_template, me.template(me.options));

    $.addClass(root, _isMounting);

    $.each(_suffixes, function (key) {
      me[key] = $.find(root, '.' + _name + '__' + key);
    });
  }

  function buildSplide(resolve, reject) {
    var me = this;
    var opts = me.options;
    var cn = me.content;
    var tmp = createElement(_template, opts.template);
    var slider = tmp;

    if (!$.isElm(cn)) {
      return;
    }

    try {
      // Might be wrapped with splide-wrapper when having navigation.
      // Splidebox currently does not support slider with navigation.
      if ($.hasClass(tmp, _base + '-wrapper')) {
        slider = $.find(tmp, '> ' + _sBase);
      }

      $.removeClass(slider, _base + '--main');
      $.attr(slider, 'data-' + _base, JSON.stringify(opts.dataSplide));
    }
    catch (e) {
      // Ignore.
    }

    append(cn, slider);

    if (opts.dataSplide.pagination) {
      $.addClass(me.root, _isPaginated);
    }

    var list = me.list = $.find(cn, '.' + _classSplideList);
    me.slider = $.find(cn, _sBase);

    if (!$.isElm(list)) {
      reject();
    }
    else {
      list.innerHTML = '';

      $.each(me.items, function (item) {
        me.addSlide(item);
      });

      resolve(list);
    }
  }

  function initSplide(list) {
    var me = this;
    var opts = me.options;
    var el = me.slider;
    var s = {};
    var root = me.root;
    var index = me.index;
    var single = me.count === 1;

    s = $.extend({}, opts.dataSplide || {}, {
      zoomRoot: root
    });

    if (single) {
      $.addClass(root, 'is-sbox-unsplide');
    }

    if (opts.init) {
      s.start = index;
      s.count = me.count;

      // el.dataset.splide = JSON.stringify(s, getCircularReplacer());
      me.splide = fnSplide = opts.init(el, s);
    }

    _isInitial = _isInitial || $.hasClass(root, _isMounting);

    setTimeout(function () {
      if (fnSplide) {
        // fnSplide.options = s;
        fnSplide.go(index);
        me.curr = fnSplide.Components.Slides.getAt(index);
        me.currData = me.items[index];

        buildPreloader.call(me);

        fnSplide.on('active.' + _nick, function (slide) {
          me.index = fnSplide.index;
          me.curr = slide;
          me.currData = me.items[me.index];

          initPreloader.call(me);
          setCaption.call(me);

          $.attr(root, _dataMedia, me.currData.type);
          $.attr(root, _dataProvider, me.currData.provider || null);

          $[me.currData.ajax ? _addClass : _removeClass](root, _isAjaxed);
          $[me.currData.display === _playable ? _addClass : _removeClass](root, _isPlayable);
        });

        fnSplide.on('moved.' + _nick, function () {
          clearDragMarkers.call(me);
        });

        // After captions are loaded.
        // @todo use an event after `active` fire instead, if any.
        setTimeout(function () {
          if (single) {
            fnSplide.destroy();
          }
        }, 100);

      }
    });
    return fnSplide;
  }

  function buildPreloader() {
    var me = this;
    var root = me.root;

    // if (!_isInitial) {
    // return;
    // }

    var data = me.currBox = me.toBox();
    if (data && !data.hasImg) {
      return;
    }

    data.initial = true;

    setClone(me, data);

    me.zoom('off');
    var clone = me.clone;

    if (!clone) {
      me.shouldPreload = false;
      return;
    }

    var reset = function () {
      var _style = clone.style;
      _style.transform = 'translate3d(0, 0, 0) scale3d(1, 1, 1)';
      _style.opacity = 0;
      $.removeClass(root, _isMounting);

      _isInitial = false;

      $.off(clone, _transitionend, reset);
      $.remove(clone);

      me.clone = null;
      me.zoom('on');
      $.addClass(root, _isActive);
    };

    $.on(clone, _transitionend, reset);

    setTimeout(function () {
      reset();
    }, 600);

    me.shouldPreload = true;
  }

  function initPreloader() {
    var me = this;
    var root = me.root;

    // if (!_isInitial) {
    // return;
    // }

    setTimeout(function () {
      var box = me.currBox;
      var img = box.img.el;

      if (box.isValid && $.isElm(img)) {
        if ($.equal(img, 'img')) {
          if (img.complete) {
            transformPreloader.call(me);
          }
          else {
            $.on(img, 'load', transformPreloader.bind(me));
          }
        }
      }
      else {
        $.removeClass(root, _isMounting);
        $.addClass(root, _isActive);
      }
    });
  }

  function transformPreloader() {
    var me = this;
    var ws = $.windowSize();
    var box = me.currBox;
    var img = box.img.el;
    var ox = 1;
    var oy = 1;
    var margin = 0;
    var sub = 0.5;
    var subWin = 2;
    var rect = $.rect(img);
    var iw = $.outerWidth(img, true);
    var ih = $.outerHeight(img, true);

    box = $.extend({}, box, {
      initial: false,
      img: {
        el: img,
        width: iw,
        height: ih,
        left: rect.left,
        top: rect.top,
        rect: rect
      }
    });

    setClone(me, box);

    var item = box.img;
    var aox = Math.abs(ox);
    var aoy = Math.abs(oy);
    var ix =
      (ox > 0 ? 1 - aox : aox) * ws.width +
      (ox * ws.width) / subWin -
      item.left -
      sub * iw;
    var iy =
      (oy > 0 ? 1 - aoy : aoy) * ws.height +
      (oy * ws.height) / subWin -
      item.top -
      sub * ih;
    var iz = Math.min(
      Math.min(ws.width * aox - margin, item.width - margin) / iw,
      Math.min(ws.height * aoy - margin, item.height - margin) / ih
    );

    ix += 4;

    me.clone.style.transform =
      'translate3d(' +
      ix +
      'px, ' +
      iy +
      'px, 0) scale3d(' +
      iz +
      ', ' +
      iz +
      ', 1)';
  }

  function attachRoot(splide) {
    var me = this;
    var root = me.root;
    var docFrag = _doc.createDocumentFragment();

    append(docFrag, root);
    append(_doc.body, docFrag);

    setTimeout(function () {
      initPlugins.call(me);
      var splide = me.splide;

      if (splide) {
        splide.refresh();
      }
    });
  }

  function buildOut() {
    var me = this;
    var opts = me.options;

    if (!me.count || !opts.template) {
      return;
    }

    buildRoot.call(me);

    var promise = new Promise(buildSplide.bind(me));
    promise.then(initSplide.bind(me));
    promise.then(attachRoot.bind(me));
  }

  function toggleCaption(e) {
    $.toggleClass(e.target, _isVisible);
  }

  function onPlaying() {
    $.addClass(this.root, _isPlaying);
  }

  function onStopped() {
    $.removeClass(this.root, _isPlaying);
  }

  function initEvents() {
    var me = this;
    var die = me.close.bind(me);

    $.on(me.closer, _click, die);
    $.on(me.overlay, _click, die);
    $.on(me.caption, _click, toggleCaption);

    $.on(_win, 'blazy.mediaPlaying', onPlaying.bind(me));
    $.on(_win, 'blazy.mediaStopped', onStopped.bind(me));
  }

  function initPlugins() {
    var me = this;
    var splide = me.splide;

    if (splide && splide.Components) {
      var components = splide.Components;

      if (!fnFullscreen && 'FullScreen' in components) {
        fnFullscreen = components.FullScreen;
        var fsOptions = {
          element: me.root,
          className: _isFullscreened
        };
        fnFullscreen.init(fsOptions);
      }

      if (!fnMedia && 'Media' in components) {
        fnMedia = components.Media;
      }

      if (!fnZoom && 'Zoom' in components) {
        fnZoom = components.Zoom;
      }
    }
  }

  function createElement(tagName, html, className, root, attach) {
    attach = attach || '';
    var el = $.create(tagName || 'div', className, html);

    // @todo replace with dBlazy post Blazy 2.6.
    if (root && attach) {
      if (attach === 'prepend') {
        root.insertBefore(el, root.childNodes[0] || null);
      }
      else if (attach === 'append') {
        append(root, el);
      }
      else if (attach === 'next') {
        root.insertAdjacentElement('afterend', el);
      }
    }
    return el;
  }

  function clearDragMarkers() {
    var root = this.root;

    setTimeout(function () {
      $.removeAttr(root, _data + 'scale');

      var moves = $.findAll(root, '.' + _dragObj);
      if (moves.length) {
        $.each(
          moves,
          function (item) {
            item.style.transform = '';

            $.removeClass(item, _isMoved);
          },
          100
        );
      }
    });
  }

  function setClone(me, data) {
    var clone = me.clone;
    var img = data.img;
    var tn = data.tn;
    var height = img.height;
    var width = img.width;
    var left = img.left;
    var top = img.top;

    if (data.initial || !clone) {
      clone = createElement('div', null, _name + '__clone', me.overlay, 'next');
      height = tn.height;
      width = tn.width;
      left = tn.left;
      top = tn.top;
    }

    if (clone) {
      var _style = clone.style;
      _style.height = height + 'px';
      _style.width = width + 'px';
      _style.left = left + 'px';
      _style.top = top + 'px';

      _style.backgroundImage = 'url(' + tn.src + ')';
      me.clone = clone;
    }
  }

  function setCaption() {
    var me = this;
    var opts = me.options;
    var el = me.caption;
    var data = me.currData;
    var captioned = $.isObj(data) &&
      data.caption &&
      !data.ajax;

    if ($.isElm(el) && opts.captionPos === _overlay) {
      el.innerHTML = captioned ? _sanitizer.sanitize(data.caption) : '';
      $[captioned ? _addClass : _removeClass](el, _isCaptioned);
    }

    if (me.counter) {
      me.counter.innerText = me.index + 1 + '/' + me.splide.length;
    }
  }

  function preload(data) {
    var _ajax = data.ajax ? '' : ' ' + _isZoomable;
    var image = createElement(_img, null, _name + '__img' + _ajax);
    var attrs = ['alt', _src, _srcset, _sizes, _height, _width];

    image.decoding = 'async';
    image.loading = 'lazy';

    $.each(attrs, function (attr) {
      if (attr in data && data[attr]) {
        image[attr] = data[attr];
      }
    });

    return image;
  }

  function stopVideo() {
    if (fnMedia) {
      fnMedia.close();
    }
  }

  function extractAttr(html, attr) {
    return html
      .split(attr + '="')
      .pop()
      .split('"')[0];
  }

  function launch(index) {
    var me = this;
    var root;

    me.index = index;

    buildOut.call(me);
    initEvents.call(me);

    root = me.root;

    if ($.isElm(root)) {
      $.addClass(root, _isMounted);
      $.attr(root, _ariaHidden, false);
    }
  }

  function detach(me) {
    var splide = me.splide;
    var root = me.root;

    if (splide) {
      stopVideo.call(me);

      fnFullscreen = false;
      fnMedia = false;
      fnZoom = false;
      splide.destroy(true);
      splide = null;
    }

    if ($.isElm(root)) {
      $.attr(root, _ariaHidden, true);
      $.remove(root);
      root = null;
    }
  }

  function append(parent, el) {
    if ($.isElm(parent)) {
      if ($.isElm(el)) {
        parent.appendChild(el);
      }
      else {
        $.append(parent, el);
      }
    }
  }

  function resetInitials(me) {
    me.root = null;

    $.each(_suffixes, function (key) {
      me[key] = null;
    });

    me.index = 0;
    me.count = 0;
    me.items = [];
    me.list = null;
    me.slider = null;
    me.triggers = [];
    me.shouldPreload = false;
    me.splide = null;
    me.curr = null;
    me.currBox = {
      img: {},
      tn: {}
    };
    me.currData = {};
  }

  function init(me) {
    resetInitials(me);

    _trigger = me.options.trigger;

    var gallery = me.gallery;
    if (gallery && $.isElm(gallery)) {
      $.on(gallery, _click, _trigger, me.open.bind(me), false);
    }
  }

  /**
   * Initializing the HTML element into a Splidebox.
   *
   * @param {HTMLElement} el
   *   The Splidebox HTML element.
   * @param {object} opts
   *   The Splidebox options.
   *
   * @return {function}
   *   The Splidebox instance.
   */
  Splidebox.init = function (el, opts) {
    if (!el.splidebox) {
      el.splidebox = new Splidebox(el, opts);
    }
    return el.splidebox;
  };

  return Splidebox;

}));
